import pathlib
import shutil

from loguru import logger

from rawfinder.finders import JpegFinder, RawFinder
from rawfinder.indexers import FileStorage


class App:
    DEFAULT_DST_FOLDER = pathlib.Path("raw")

    def __init__(
        self,
        jpeg_images_path: pathlib.Path,
        raw_images_path: pathlib.Path,
        raw_images_dest_path: pathlib.Path,
    ):
        self.jpeg_finder = JpegFinder(jpeg_images_path)
        self.raw_finder = RawFinder(raw_images_path)
        self.raw_images_dest_path = raw_images_dest_path

        self.raw_images_dest_path = (
            raw_images_dest_path if raw_images_dest_path else jpeg_images_path / self.DEFAULT_DST_FOLDER
        )

    def get_user_confirmation(self) -> None:
        """
        Prompts the user for confirmation to proceed.
        """
        message = (
            f"JPEGs folder: '{self.jpeg_finder.path}'\n"
            f"RAWs folder: '{self.raw_finder.path}'\n"
            f"Destination folder: '{self.raw_images_dest_path}'\n"
            "This script will find corresponding RAW files for these JPEG files and copy them to the destination folder.\n"
            "Is it ok: [Y/n] "
        )

        if input(message).lower() not in ["y", ""]:
            raise KeyboardInterrupt("Cancelled.")

    def prepare_destination(self) -> None:
        logger.info(f"Creating destination folder: {self.raw_images_dest_path}")
        self.raw_images_dest_path.mkdir(exist_ok=True, parents=True)

    def process_files(self) -> None:
        logger.debug("Indexing RAW files")

        storage = FileStorage()
        storage.make_index(self.raw_finder.find())

        logger.debug("Processing JPEG files")

        for jpeg_file in self.jpeg_finder.find():
            raw_file = storage.get(jpeg_file.stem.lower())
            if raw_file:
                logger.info(
                    f"RAW file {raw_file.name} found for {jpeg_file.name}, copying to {self.raw_images_dest_path}..."
                )
                shutil.copy(raw_file, self.raw_images_dest_path)
            else:
                logger.warning(f"No RAW file found for {jpeg_file.name}!")

    def start(self) -> None:
        """
        Starts the application workflow.
        """
        try:
            self.get_user_confirmation()
            self.prepare_destination()
            self.process_files()
            logger.info("Done.")
        except KeyboardInterrupt:
            pass



import pathlib


class FileStorage:
    def __init__(self) -> None:
        """Initializes an empty storage dictionary and a flag to track if indexing is complete."""

        self.indexed = False
        self._storage: dict[str, pathlib.Path] = {}

    def make_index(self, files: list[pathlib.Path]) -> None:
        """Populates the internal storage with the provided list of file paths.
        Each file's stem (filename without extension) is converted to lowercase
        and used as a key, while the full path is stored as the value.
        """
        for file in files:
            self._storage[file.stem.lower()] = file

        # Mark the storage as indexed
        self.indexed = True

    def get(self, file_name: str) -> pathlib.Path | None:
        """Retrieves a file path by its name, ignoring case."""

        if not self.indexed:
            raise ValueError()

        return self._storage.get(file_name.lower())



import pathlib
import typing
from functools import lru_cache


class BaseFinder:
    extensions: typing.ClassVar[list[str]] = []

    def __init__(self, path: pathlib.Path) -> None:
        self.path = path

    @classmethod
    @lru_cache
    def _convert_to_case_insensitive(cls, extension: str) -> str:
        """
        The case_sensitive param for glob/rglob was added in Python 3.12
        this method allow emulation of case_sensitive for older versions
        of Python
        """
        ext = "".join(f"[{c.lower()}{c.upper()}]" for c in extension[1:])
        return f"*{extension[0]}{ext}"

    def find(self) -> list[pathlib.Path]:
        return [
            file
            for ext in self.extensions
            for file in pathlib.Path(self.path).glob(
                BaseFinder._convert_to_case_insensitive(ext),
            )
        ]

    def __str__(self) -> str:
        return str(self.path)


class JpegFinder(BaseFinder):
    extensions: typing.ClassVar[list[str]] = [".jpeg", ".jpg"]


class RawFinder(BaseFinder):
    extensions: typing.ClassVar[list[str]] = [
        ".3fr",  # Hasselblad
        ".ari",  # Arri Alexa
        ".arw",
        ".srf",
        ".sr2",  # Sony
        ".bay",  # Casio
        ".braw",  # Blackmagic Design
        ".cri",  # Cintel
        ".crw",
        ".cr2",
        ".cr3",  # Canon
        ".cap",
        ".iiq",
        ".eip",  # Phase One
        ".dcs",
        ".dcr",
        ".drf",
        ".k25",
        ".kdc",  # Kodak
        ".dng",  # Adobe
        ".erf",  # Epson
        ".fff",  # Imacon/Hasselblad raw
        ".gpr",  # GoPro
        ".mef",  # Mamiya
        ".mdc",  # Minolta, Agfa
        ".mos",  # Leaf
        ".mrw",  # Minolta, Konica Minolta
        ".nef",
        ".nrw",  # Nikon
        ".orf",  # Olympus
        ".pef",
        ".ptx",  # Pentax
        ".pxn",  # Logitech
        ".R3D",  # RED Digital Cinema
        ".raf",  # Fuji
        ".raw",
        ".rw2",  # Panasonic
        ".raw",
        ".rwl",
        ".dng",  # Leica
        ".rwz",  # Rawzor
        ".srw",  # Samsung
        ".tco",  # intoPIX
        ".x3f",  # Sigma
    ]



import pathlib

import click

from rawfinder.app import App


@click.command()
@click.argument("images_dir", type=click.Path(exists=True))
@click.argument("sources_dir", type=click.Path(exists=True))
@click.argument("dest_sources_dir", default="", type=click.Path(writable=True))
def main(images_dir: pathlib.Path, sources_dir: pathlib.Path, dest_sources_dir: pathlib.Path) -> None:
    """Find corresponding RAW files for JPEG images and copy them to a DEST folder.

    \b
    JPEG_DIR - directory with JPEG files.
    RAW_DIR  - directory with RAW files.
    DEST_DIR - destination directory for RAW files.
               default is 'raw' inside the JPEG_DIR
    """

    app = App(
        images_dir,
        sources_dir,
        dest_sources_dir,
    )
    app.start()


if __name__ == "__main__":
    main()
